﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using Infragistics.Controls.Charts;

namespace Infragistics.Extension
{
    /// <summary>
    /// Represents a custom type of CustomScatterSeries for XamDataChart control.
    /// </summary>
    public class CustomScatterSeries : Series //MarkerSeries //Series
    {
        public CustomScatterSeries()
        {
            this.DefaultStyleKey = typeof(CustomScatterSeries);
        }
        
        protected KDTree<object> ItemsTree;
        
        #region Layers Support

        /// <summary>
        /// If possible, will return the best available position within the series that has the best value match for the world position provided. 
        /// </summary>
        /// <param name="world">The world coordinate for which to get a value position for</param>
        /// <param name="useInterpolation">If true, interpolation should be used to get in-between values, rather than only the actual values in the data set.</param>
        /// <param name="skipUnknowns">If true, unknown values will be skipped when searching for a series value.</param>
        /// <returns>The position in question, or NaN,NaN if no valid position is available.</returns>
        public override Point GetSeriesValuePosition(Point world, bool useInterpolation, bool skipUnknowns)
        {
            var seriesPoint = new Point(double.NaN, double.NaN);

            if (!world.IsPlottable()) return seriesPoint;

            var position = GetSeriesExactPosition(world);
            var neighbors = this.ItemsTree.GetNeighbors(3, position.X, position.Y).ToList();
           
            if (neighbors.Any())
            {
                Point closestPoint;
  
                var itemA = neighbors[0];
                var pointA = new Point(itemA.Point[0], itemA.Point[1]);

                if (useInterpolation && neighbors.Count() >= 2)
                {
                    var itemB = neighbors[1];
                    var pointB = new Point(itemB.Point[0], itemB.Point[1]);
                    closestPoint = position.GetIntersectionWith(pointA, pointB);

                    if (!closestPoint.IsOnSegment(pointA, pointB))
                    {
                        closestPoint = pointA;
                    }
                }
                else
                {
                    closestPoint = pointA; 
                }

                seriesPoint = GetSeriesScaledPosition(closestPoint);
               
                //var neighborString = "";
                //foreach (var item in neighbors)
                //{
                //    neighborString += item.Point[0] + "," + item.Point[1] + ",(" + item.Distance + ") | ";
                //}
                //Debug.WriteLine("Tracker " + (int)closestPoint.X + " " + (int)closestPoint.Y +
                //                   " p=" + position.X + ", " + position.Y +
                //                   " w=" + world.X + ", " + world.Y +
                //                   "\n" + neighborString);

            }
           
            return seriesPoint;

        }

        /// <summary>
        /// Gets the item item associated with the specified world position 
        /// where a tooltip must be displayed on the series' marker
        /// </summary>
        public override object GetItem(Point world)
        {
            //var seriesPoint = new Point(double.NaN, double.NaN);

            var position = GetSeriesExactPosition(world);
            var neighbors = this.ItemsTree.GetNeighbors(2, position.X, position.Y);

            var closest = neighbors.FirstOrDefault();
            if (closest != null)
            {
                // var x = closest.Point[0];
                //var y = closest.Point[1];
                var item = closest.Value;
                return item;
                //seriesPoint = GetSeriesScaledPosition(new Point(x, y));
            }


            return null;
        }
       
        #endregion


        public override bool IsScatter { get  { return true; } }
        public override bool IsCategory { get { return false; } }
        

        #region Properties

        internal const string ItemsMemberPathXPropertyName = "ItemsMemberPathX";
        public static readonly DependencyProperty ItemsMemberPathXProperty = DependencyProperty.Register(
            ItemsMemberPathXPropertyName, typeof(string), typeof(CustomScatterSeries),
            new PropertyMetadata("X", (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(ItemsMemberPathXPropertyName, e.OldValue, e.NewValue);
            }));
        /// <summary>
        /// Gets or sets x-value member path for data items
        /// </summary>
        public string ItemsMemberPathX
        {
            get { return (string)GetValue(ItemsMemberPathXProperty); }
            set { SetValue(ItemsMemberPathXProperty, value); }
        }

        internal const string ItemsMemberPathYPropertyName = "ItemsMemberPathY";
        public static readonly DependencyProperty ItemsMemberPathYProperty = DependencyProperty.Register(
            ItemsMemberPathYPropertyName, typeof(string), typeof(CustomScatterSeries),
            new PropertyMetadata("Y", (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(ItemsMemberPathYPropertyName, e.OldValue, e.NewValue);
            }));
        /// <summary>
        /// Gets or sets y-value member path for data items
        /// </summary>
        public string ItemsMemberPathY
        {
            get { return (string)GetValue(ItemsMemberPathYProperty); }
            set { SetValue(ItemsMemberPathYProperty, value); }
        }

        internal const string ItemsMemberPathRPropertyName = "ItemsMemberPathR";
        public static readonly DependencyProperty ItemsMemberPathRProperty = DependencyProperty.Register(
            ItemsMemberPathRPropertyName, typeof(string), typeof(CustomScatterSeries),
            new PropertyMetadata("Value", (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(ItemsMemberPathRPropertyName, e.OldValue, e.NewValue);
            }));

        /// <summary>
        /// Gets or sets r-value member path for data items
        /// </summary>
        public string ItemsMemberPathR
        {
            get { return (string)GetValue(ItemsMemberPathRProperty); }
            set { SetValue(ItemsMemberPathRProperty, value); }
        }

        #region Property - XAxis
        public const string XAxisPropertyName = "XAxis";
        public static readonly DependencyProperty XAxisProperty =
            DependencyProperty.Register(XAxisPropertyName, typeof(NumericXAxis),
            typeof(CustomScatterSeries), new PropertyMetadata(null, (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(XAxisPropertyName, e.OldValue, e.NewValue);
            }));
        /// <summary>
        /// Gets or sets x-axis 
        /// </summary>
        public NumericXAxis XAxis
        {
            get { return this.GetValue(XAxisProperty) as NumericXAxis; }
            set { this.SetValue(XAxisProperty, value); }
        }
        #endregion
        #region Property - YAxis
        public const string YAxisPropertyName = "YAxis";
        public static readonly DependencyProperty YAxisProperty =
            DependencyProperty.Register(YAxisPropertyName, typeof(NumericYAxis),
            typeof(CustomScatterSeries), new PropertyMetadata(null, (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(YAxisPropertyName, e.OldValue, e.NewValue);
            }));
        /// <summary>
        /// Gets or sets y-axis 
        /// </summary>
        public NumericYAxis YAxis
        {
            get { return this.GetValue(YAxisProperty) as NumericYAxis; }
            set { this.SetValue(YAxisProperty, value); }
        }
        #endregion

        public BrushCollection ActualContourBrushes { get; private set; }
        public BrushCollection ActualContourOutlines { get; private set; }

        #endregion
   
        #region Brush Methods
        private Brush GetSeriesFill(int conturIndex)
        {
            return GetValidBrush(conturIndex, ((XamDataChart)this.SeriesViewer).Brushes);
        }
        private Brush GetSeriesStroke(int conturIndex)
        {
            return GetValidBrush(conturIndex, ((XamDataChart)this.SeriesViewer).MarkerOutlines);
        }
        private Brush GetMarkerOutline(int conturIndex)
        {
            return GetValidBrush(conturIndex, ((XamDataChart)this.SeriesViewer).MarkerOutlines);
        }
        private Brush GetMarkerFill(int conturIndex)
        {
            return GetValidBrush(conturIndex, ((XamDataChart)this.SeriesViewer).MarkerBrushes);
        }
        private static Brush GetValidBrush(int conturIndex, BrushCollection brushes)
        {
            if (brushes == null || brushes.Count == 0)
            {
                return new SolidColorBrush(Colors.Black);
            }
            if (conturIndex >= 0 && conturIndex < brushes.Count)
            {
                return brushes[conturIndex];
            }
            conturIndex = conturIndex % brushes.Count;
            return brushes[conturIndex];
        }
        #endregion

        #region Overriden Methods
        /// <summary>
        /// Renders the Custom Series
        /// </summary>
        protected override void RenderSeriesOverride(bool animate)
        {
            base.RenderSeriesOverride(animate);
            // check if the series can be rendered:
            // - the Viewport (the bounds rectangle for the series) is not empty, 
            // - the RootCanvas (the container for the custom graphics) is not null.  
            // - the Axes are not null.  
            // - the ItemsSource is not null.  
            if (this.Viewport.IsEmpty || this.RootCanvas == null ||
                this.XAxis == null || this.YAxis == null ||
                this.ItemsSource == null)
            {
                return;
            }
            // clears the RootCanvas on every render of the series
            this.RootCanvas.Children.Clear();

            // create data structure for the Custom Series based on values of items in the source of this series
            var itemsDictionary = new Dictionary<double, PointCollection>();
            foreach (var dataItem in this.ItemsSource)
            {
                var dataPoint = GetDataPoint(dataItem);
                var isInverted = this.XAxis.IsInverted;
                var param = new ScalerParams(this.SeriesViewer.WindowRect, this.Viewport, isInverted);
                // scale x-locations of data point to the series' viewport
                var x = this.XAxis.GetScaledValue(dataPoint.X, param);

                isInverted = this.YAxis.IsInverted;
                param = new ScalerParams(this.SeriesViewer.WindowRect, this.Viewport, isInverted);
                // scale y-location of data point to the series' viewport
                var y = this.YAxis.GetScaledValue(dataPoint.Y, param);

                // store scaled locations of data point based on the Value property of data points
                var key = dataPoint.Value;
                if (itemsDictionary.ContainsKey(key))
                {
                    itemsDictionary[key].Add(new Point(x, y));
                }
                else
                {
                    var dataPoints = new PointCollection { new Point(x, y) };
                    itemsDictionary.Add(key, dataPoints);
                }
            }
            // sort data items based on key (value of data point)
            var itemsSorted = from item in itemsDictionary
                                 orderby item.Key ascending
                                 select item;

            
            // create elements of markers based on data structure
            var index = 0;
            foreach (var item in itemsSorted) //dataContours)
            {
                foreach (var point in item.Value)
                {
                    // get parameters of a marker
                    var markerValue = item.Key;
                    var markerSize = 25; //item.Key; //25;
                    var markerLocationLeft = point.X - markerSize / 2.0;
                    var markerLocationTop = point.Y - markerSize / 2.0;

                    // create element for shape of a marker
                    var marker = new Ellipse();
                    marker.Fill = GetMarkerFill(index);
                    marker.Stroke = GetMarkerOutline(index);
                    marker.StrokeThickness = this.Thickness;
                    marker.Width = markerSize;
                    marker.Height = markerSize;

                    // create element for value of a marker
                    var markerValueBlock = new TextBlock();
                    markerValueBlock.Text = markerValue.ToString();
                    markerValueBlock.Foreground = new SolidColorBrush(Colors.White);
                    markerValueBlock.VerticalAlignment = VerticalAlignment.Center;
                    markerValueBlock.HorizontalAlignment = HorizontalAlignment.Center;

                    // create element to hold elements of a marker 
                    var markerGrid = new Grid();
                    markerGrid.Children.Add(marker);
                    markerGrid.Children.Add(markerValueBlock);
                    Canvas.SetLeft(markerGrid, markerLocationLeft);
                    Canvas.SetTop(markerGrid, markerLocationTop);
                    Canvas.SetZIndex(markerGrid, index + 11);

                    // render the marker of the data item on the canvas of this series
                    this.RootCanvas.Children.Add(markerGrid);
                }

                //OPTIONAL 
                //var contourPoints = contour.Value;
                //// create curve from points of a contour
                ////var contourFigure = ShapeBuilder.GetBezierSegments(contourPoints, 1.0, true);
                //var contourFigure = ShapeBuilder.GetBezierSegments(contourPoints,8.4, true);
                //contourFigure.IsClosed = true;

                //// create a new PathGeometry for a contour
                //var contourGeo = new PathGeometry();
                //contourGeo.Figures.Add(contourFigure);
                //// create a new Path for a contour
                //var contourShape = new Path();
                //contourShape.Data = contourGeo;
                //contourShape.Stroke = GetContourPathStroke(conturIndex);
                //contourShape.StrokeThickness = this.Thickness;
                //contourShape.Fill = GetContourPathFill(conturIndex);
                //Canvas.SetZIndex(contourShape, conturIndex + 10);

                //// render shape of the current contour on the canvas of this series
                //this.RootCanvas.Children.Add(contourShape);
                index++;
            }
        }

        /// <summary>
        /// Redraws series any time the Viewport rect changes
        /// </summary>
        protected override void ViewportRectChangedOverride(Rect oldViewportRect, Rect newViewportRect)
        {
            base.ViewportRectChangedOverride(oldViewportRect, newViewportRect);
            this.RenderSeries(false);
        }
        /// <summary>
        /// Redraws series any time the Window rect changes
        /// </summary>
        protected override void WindowRectChangedOverride(Rect oldWindowRect, Rect newWindowRect)
        {
            base.WindowRectChangedOverride(oldWindowRect, newWindowRect);
            this.RenderSeries(false);
        }

        /// <summary>
        /// Overrides property update of this series
        /// </summary>
        protected override void PropertyUpdatedOverride(object sender, string propertyName, object oldValue, object newValue)
        {
            base.PropertyUpdatedOverride(sender, propertyName, oldValue, newValue);
            switch (propertyName)
            {
                case ItemsSourcePropertyName:
                case ItemsMemberPathXPropertyName:
                case ItemsMemberPathYPropertyName:
                case ItemsMemberPathRPropertyName:

                    //var dataPoints = GetDataPoints();
                    //var dataType = this.ItemsSource

                    this.ItemsTree = new KDTree<object>(2);
                    foreach (var item in this.ItemsSource)
                    {
                        var dataItem = GetDataPoint(item);

                        this.ItemsTree.Add(item, dataItem.X, dataItem.Y);
                    }
                    
                    this.RenderSeries(false);
                    if (this.XAxis != null)
                    {
                        this.XAxis.UpdateRange();
                    }
                    if (this.YAxis != null)
                    {
                        this.YAxis.UpdateRange();
                    }
                    break;

                case XAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }

                    if ((XAxis != null && !XAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;

                case YAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }
                    if ((YAxis != null && !YAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;
            }
        }

        /// <summary>
        /// Calculates range of a given axis based on  X/Y values of data items
        /// </summary>
       protected override AxisRange GetRange(Axis axis)
        {
            if (this.ItemsSource == null)
            {
                return base.GetRange(axis);
            }
            if (axis == this.XAxis)
            {
                var min = double.MaxValue;
                var max = double.MinValue;
                foreach (var dataItem in this.ItemsSource)
                {
                    var dataPoint = GetDataPoint(dataItem);
                    min = Math.Min(min, dataPoint.X);
                    max = Math.Max(max, dataPoint.X);
                }
                return new AxisRange(min, max);
            }
            if (axis == this.YAxis)
            {
                var min = double.MaxValue;
                var max = double.MinValue;
                foreach (var dataItem in this.ItemsSource)
                {
                    var DataPoint = GetDataPoint(dataItem);
                    min = Math.Min(min, DataPoint.Y);
                    max = Math.Max(max, DataPoint.Y);
                }
                return new AxisRange(min, max);
            }
            
            return base.GetRange(axis);
        }
        
    
        #endregion

        #region Help Methods
        
        public Point GetSeriesScaledPosition(Point dataPoint)
       {
           var windowRect = this.SeriesViewer.ActualWindowRect;
           var viewport = this.Viewport;

           var xParams = new ScalerParams(windowRect, viewport, this.XAxis.IsInverted);
           xParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var yParams = new ScalerParams(windowRect, viewport, this.YAxis.IsInverted);
           yParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var x = this.XAxis.GetScaledValue(dataPoint.X, xParams);
           var y = this.YAxis.GetScaledValue(dataPoint.Y, yParams);

           return new Point(x, y);
       }
      
        public Point GetSeriesExactPosition(Point world)
       {
           var windowRect = this.SeriesViewer.ActualWindowRect;
           var viewport = this.Viewport;

           var xParams = new ScalerParams(windowRect, viewport, this.XAxis.IsInverted);
           xParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var yParams = new ScalerParams(windowRect, viewport, this.YAxis.IsInverted);
           yParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           double x;
           if (this.XAxis.IsVertical)
           {
               x = world.Y;
               x = viewport.Top + viewport.Height * (x - windowRect.Top) / windowRect.Height;
           }
           else
           {
               x = world.X;
               x = viewport.Left + viewport.Width * (x - windowRect.Left) / windowRect.Width;
           }
           double y;
           if (this.YAxis.IsVertical)
           {
               y = world.Y;
               y = viewport.Top + viewport.Height * (y - windowRect.Top) / windowRect.Height;
           }
           else
           {
               y = world.X;
               y = viewport.Left + viewport.Width * (y - windowRect.Left) / windowRect.Width;
           }

           x = this.XAxis.GetUnscaledValue(x, xParams);
           y = this.YAxis.GetUnscaledValue(y, yParams);


           return new Point(x, y);
       }
     
        private DataPoint GetDataPoint(object dataItem)
        {
            var x = (double)dataItem.GetType().GetProperty(this.ItemsMemberPathX).GetValue(dataItem, null);
            var y = (double)dataItem.GetType().GetProperty(this.ItemsMemberPathY).GetValue(dataItem, null);
            var v = (double)dataItem.GetType().GetProperty(this.ItemsMemberPathR).GetValue(dataItem, null);
            return new DataPoint(x, y, v);
        }

        //private IEnumerable<DataPoint> GetDataPoints()
        //{
        //    var list = new List<DataPoint>();
        //    foreach (var item in this.ItemsSource)
        //    {
        //        var point = GetDataPoint(item);
        //        list.Add(point);
        //    }
        //    return list;
        //} 
        #endregion

        //public delegate void SeriesMoveChangedEventHandler(object sender, SeriesMoveEventArgs e);
        //public event SeriesMoveChangedEventHandler SeriesMove;

    }

    internal class DataPoint
    {
        public DataPoint()
        {
            
        }

        public DataPoint(double x, double y,double v)
        {
            X = x;
            Y = y;
            Value = v;
        }
        /// <summary>
        /// Gets or sets X
        /// </summary>
        public double X { get; set; }

        /// <summary>
        /// Gets or sets Y
        /// </summary>
        public double Y { get; set; }

        /// <summary>
        /// Gets or sets Point
        /// </summary>
        public Point Point { get { return new Point(X, Y);} }
        /// <summary>
        /// Gets or sets Value
        /// </summary>
        public double Value { get; set; }

    }


    //public class SeriesMoveEventArgs : EventArgs
    //{
    //    public object DataItem { get; set; }
    //    public Point DataPoint { get; set; }
    //    public Point WorldPoint { get; set; }


    //    public SeriesMoveEventArgs(Point worldPoint, Point dataPoint, object dataItem)
    //    {
    //        this.DataItem = dataItem;
    //        this.DataPoint = dataPoint;
    //        this.WorldPoint = worldPoint;
    //    }
    //}

}